// DocFileUtils.cpp

#include "StdAfx.h"

#include "Common/IntToString.h"
#include "Common/StringToInt.h"

#include "DocFileUtils.h"

static const UString kFileSysUnallowedChars = L"?*<>";

static const UInt32 specialCharCodeMax = 0x20;

static const wchar_t kCharOpenBracket  = L'[';
static const wchar_t kCharCloseBracket = L']';

static UString UInt32ToString(UInt32 value)
{
  wchar_t temp[32];
  ConvertUInt64ToString(value, temp);
  return temp;
}

static bool StringToUInt32(const UString &value, UInt32 &result)
{
  UInt64 res64;
  const WCHAR *end;
  res64 = ConvertStringToUInt64(value, &end);
  if (*end != 0)
    return false;
  result = (UInt32)res64;
  return (res64 < ((UInt64)1 << 32));
}

UString CompoundNameToFileName(const UString &name, bool fileSysMode)
{
  UString result;
  for (int i = 0; i < name.Length(); i++)
  {
    wchar_t c = name[i];
    if (c < specialCharCodeMax || (fileSysMode && kFileSysUnallowedChars.Find(c) >= 0))
    {
      result += kCharOpenBracket;
      result += UInt32ToString(c);
      result += kCharCloseBracket;
    }
    else
      result += c;
  }
  return result;
}

UString FileNameToCompoundName(const UString &name)
{
  UString result;
  UString numberString;
  bool numberMode = false;
  for (int i = 0; i < name.Length(); i++)
  {
    wchar_t c = name[i];
    if (numberMode)
    {
      if (c == kCharCloseBracket)
      {
        UInt32 number;
        if (StringToUInt32(numberString, number) && number < 0x40)
          result += (wchar_t)number;
        else
        {
          result += kCharOpenBracket;
          result += numberString;
          result += kCharCloseBracket;
        }
        numberMode = false;
      }
      else
        numberString += c;
    }
    else
    {
      if (c == kCharOpenBracket)
      {
        numberString.Empty();
        numberMode = true;
      }
      else
        result += c;
    }
  }
  return result;
}

static char g_MsiChars[] =
"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz._";

static int GetMsiCharValueFromChar(wchar_t c)
{
  for (int i = 0; i < 10 + 26 + 26 + 2; i++)
    if (g_MsiChars[i] == c)
      return i;
  throw "Msi: Incorrect char";
}

static const wchar_t kMsi_SpecChar = L'!';
static const wchar_t *kMsi_ID = L"{Msi}";
static const int kMsi_ID_Size = 5;

static const int kMsiNumBits = 6;
static const int kMsiNumChars = 1 << kMsiNumBits;
static const int kMsiCharMask = kMsiNumChars - 1;
static const int kMsiStartUnicodeChar = 0x3800;
static const int kMsiUnicodeRange = 0x1000 + kMsiNumChars;

bool CompoundMsiNameToFileName(const UString &name, UString &result)
{
  result.Empty();
  for (int i = 0; i < name.Length(); i++)
  {
    wchar_t c =  name[i];
    if (c < kMsiStartUnicodeChar || c > kMsiStartUnicodeChar + kMsiUnicodeRange)
      return false;
    if (i == 0)
      result += kMsi_ID;
    c -= kMsiStartUnicodeChar;
    
    int c0 = c & kMsiCharMask;
    int c1 = c >> kMsiNumBits;

    if (c1 <= kMsiNumChars)
    {
      result += (wchar_t)g_MsiChars[c0];
      if (c1 == kMsiNumChars)
        return true; // end of name
      result += (wchar_t)g_MsiChars[c1];
    }
    else
      result += kMsi_SpecChar;
  }
  return true;
}

bool FileNameToCompoundMsiName(const UString &name, UString &result)
{
  result.Empty();
  if (name.Length() <= kMsi_ID_Size || wcsncmp(name, kMsi_ID, kMsi_ID_Size) != 0)
    return false;

  for (int i = kMsi_ID_Size; i < name.Length();)
  {
    wchar_t c = name[i++];
    int c0, c1;
    if (c == kMsi_SpecChar)
    {
      c1 = kMsiNumChars + 1;
      c0 = 0;
    }
    else
    {
      c0 = GetMsiCharValueFromChar(c);
      c1 = (i < name.Length()) ?
        GetMsiCharValueFromChar(name[i++]) :
        kMsiNumChars;
    }
    result += (wchar_t)(kMsiStartUnicodeChar + (c1 << kMsiNumBits) + c0);
  }
  return true;
}
