// StorageWalk.cpp

#include "StdAfx.h"

#include "Windows/Error.h"

#include "StorageWalk.h"

struct CItsControlData
{
  UINT32 Num;
  UINT32 Data[1]; // Data[Num]
};

typedef enum EItsCompact
{
  ITS_COMPACT_DATA = 0,
  ITS_COMPACT_DATA_AND_PATH
};

DECLARE_INTERFACE_(IITStorage, IUnknown)
{
  STDMETHOD(StgCreateDocfile)(const WCHAR *name, DWORD grfMode, DWORD reserved, IStorage**storage) PURE;
  STDMETHOD(StgCreateDocfileOnILockBytes)(ILockBytes *plkbyt, DWORD grfMode, DWORD reserved, IStorage **storage) PURE;
  STDMETHOD(StgIsStorageFile) (const WCHAR *name) PURE;
  STDMETHOD(StgIsStorageILockBytes) (ILockBytes *plkbyt) PURE;
  STDMETHOD(StgOpenStorage)(const WCHAR *name, IStorage *pstgPriority,
      DWORD grfMode, SNB snbExclude, DWORD reserved, IStorage **storage) PURE;
  STDMETHOD(StgOpenStorageOnILockBytes)(ILockBytes *plkbyt, IStorage *pStgPriority, DWORD grfMode,
      SNB snbExclude, DWORD reserved, IStorage **storage) PURE;
  STDMETHOD(StgSetTimes)(WCHAR const *lpszName, FILETIME const *cTime, FILETIME const *aTime, FILETIME const *mTime) PURE;
  STDMETHOD(SetControlData)(CItsControlData *controlData) PURE;
  STDMETHOD(DefaultControlData)(CItsControlData **controlData) PURE;
  STDMETHOD(Compact)(const WCHAR *name, EItsCompact compact) PURE;
};

using namespace NWindows;

static void MyThrow(HRESULT result, const char *message)
{
  CSysString sysMessage;
  if (!NError::MyFormatMessage(result, sysMessage))
    throw message;
  sysMessage.Replace(0x0A, " ");
  sysMessage.Replace(0x0D, " ");
  throw sysMessage;
}

void CItemStat::SetFrom(const STATSTG &a)
{
  Name = a.pwcsName;
  Type = STGTY(a.type);
  Size = a.cbSize;
  MTime = a.mtime;
  CTime = a.ctime;
  ATime = a.atime;
  grfMode = a.grfMode;
  grfLocksSupported = a.grfLocksSupported;
  ClassID = a.clsid;
  grfStateBits = a.grfStateBits;
}

void COpenStorages::CloseLastStorage()
{
  if (m_Storages.Size() == 0)
    throw 42348;
  m_Storages.DeleteBack();
  m_Names.DeleteBack();
}

void COpenStorages::CloseAllUpToSpecified(int index)
{
  while (m_Storages.Size() > index)
    CloseLastStorage();
}

void COpenStorages::AddRootStorage(IStorage *newStorage)
{
  if (m_Storages.Size() != 0)
    throw 42350;
  m_Storages.Add(newStorage);
  m_Names.Add(UString());
}

HRESULT COpenStorages::CreateStorage(const WCHAR *name)
{
  if (m_Storages.Size() == 0)
    throw 42349;
  CMyComPtr<IStorage> newStorage;
  HRESULT result = m_Storages.Back()->CreateStorage(name,
      STGM_DIRECT | STGM_READWRITE | STGM_SHARE_EXCLUSIVE | STGM_FAILIFTHERE,
      0, 0, &newStorage);
  switch(result)
  {
    case S_OK:
    case STG_E_FILEALREADYEXISTS:
       break;
    default: MyThrow(result, "create storage error");
  }
  return result;
}


void COpenStorages::OpenNewStorage(const WCHAR *name)
{
  if (m_Storages.Size() == 0)
    throw 42349;
  CMyComPtr<IStorage> newStorage;
  // STGM_SHARE_DENY_WRITE
  HRESULT result = m_Storages.Back()->OpenStorage(name, NULL, GetAccessMode2(), NULL, 0, &newStorage);
  if (result != S_OK)
    MyThrow(result, "Can not open storage");
  m_Storages.Add(newStorage);
  m_Names.Add(name);
};

void COpenStorages::OpenStream(const WCHAR *name, IStream **stream)
{
  if (m_Storages.Size() == 0)
    throw 41613;
  HRESULT result = m_Storages.Back()->OpenStream(name, NULL, GetAccessMode2(), 0, stream);
  if (result != S_OK)
    MyThrow(result, "Can not open stream");
};

void COpenStorages::CreateStream(const WCHAR *name, IStream **stream)
{
  if (m_Storages.Size() == 0)
    throw 41613;
  HRESULT result = m_Storages.Back()->CreateStream(name, STGM_CREATE | GetAccessMode2(), 0, 0, stream);
  if (result != S_OK)
    MyThrow(result, "Can not create stream");
};

/*
void COpenStorages::GetStorageStat(int i, CItemStat &stat) const
{
  STATSTG stat;
  HRESULT result = m_Storages[i]->Stat(&stat, STATFLAG_DEFAULT);
  if (result != S_OK)
    MyThrow(result, "Can not Retrieves the statistical information about an open storage");
  stat.SetFrom(stat);
  CoTaskMemFree(stat.pwcsName);
}

CComBSTR COpenStorages::GetStorageName(int i) const
{
  CItemStat stat;
  GetStorageStat(i, stat);
  return stat.Name;
}
*/

void COpenStorages::OpenStorageAlways(const WCHAR *name)
{
  if (m_Storages.Size() == 0)
    throw 42349;
  CMyComPtr<IStorage> lastStorage = m_Storages.Back();
  CMyComPtr<IStorage> newStorage;
  HRESULT result = lastStorage->OpenStorage(name, NULL, GetAccessMode2(), NULL, 0, &newStorage);
  if (result == STG_E_FILENOTFOUND)
  {
    result = lastStorage->CreateStorage(name, GetAccessMode2(), 0, 0, &newStorage);
    if (result != S_OK)
      MyThrow(result, "can not create storage");
  }
  else if (result != S_OK)
    MyThrow(result, "Open storage error");
  m_Storages.Add(newStorage);
  m_Names.Add(name);
};

void COpenStorages::DestroyElement(const WCHAR *name)
{
  if (m_Storages.Size() == 0)
    throw 42349;
  HRESULT result = m_Storages.Back()->DestroyElement(name);
  if (result != S_OK)
    MyThrow(result, "Can not destroy element from storage");
}

static const GUID CLSID_ITStorage = {0x5D02926A, 0x212E, 0x11D0, {0x9D, 0xF9, 0x0, 0xA0, 0xC9, 0x22, 0xE6, 0xEC}};
static const GUID IID_ITStorage   = {0x88CC31DE, 0x27AB, 0x11D0, {0x9D, 0xF9, 0x0, 0xA0, 0xC9, 0x22, 0xE6, 0xEC}};

static HRESULT OpenRootStorage2(const WCHAR *name, IStorage **newStorage, bool readOnly)
{
  DWORD accessMode = readOnly ? STGM_READ: STGM_READWRITE;;
  DWORD accessMode2 = STGM_DIRECT | accessMode | STGM_SHARE_EXCLUSIVE;

  HRESULT result = StgOpenStorage(name, NULL, accessMode2, NULL, 0, newStorage);
  if (result == S_OK)
    return result;

  CMyComPtr<IITStorage> itStorage;
  RINOK(CoCreateInstance(CLSID_ITStorage, NULL, CLSCTX_ALL, IID_ITStorage, (void **)&itStorage));
  // TGM_READ | STGM_SHARE_DENY_WRITE
  return itStorage->StgOpenStorage (name, NULL, accessMode2, NULL, 0, newStorage);
}

HRESULT OpenRootStorage(const WCHAR *name, IStorage **newStorage, bool readOnly)
{
  HRESULT result = OpenRootStorage2(name, newStorage, readOnly);
  if (result == S_OK || readOnly)
    return result;
  return OpenRootStorage2(name, newStorage, true);
}

UInt64 GetStreamSize(IStream *stream)
{
  STATSTG stat;
  HRESULT result = stream->Stat(&stat, STATFLAG_NONAME);
  if (result != S_OK)
    MyThrow(result, "::Stat error");
  return stat.cbSize.QuadPart;
}
