// CoderUtils.cpp

#include "StdAfx.h"

#include "Common/IntToString.h"

#include "CoderUtils.h"

static const wchar_t kCharCR = 13;
static const wchar_t kCharLF = 10;
static const wchar_t kCharTAB = 9;
static const wchar_t kCharQuote = L'\"';

static const wchar_t kCharSpace = L' ';
static const wchar_t kCharComma = L',';

static const wchar_t kCharORDelimeter = L'|';

using namespace std;

void CUINT32FromToUStringMap::AddType(UINT32 number, const UString &s)
{
  m_NumberToStringMap.insert(map<UINT32, UString>::value_type(number, s));
  m_StringToNumberMap.insert(map<UString, UINT32>::value_type(s, number));
}

bool CUINT32FromToUStringMap::GetString(UINT32 number, UString &s) const
{
  map<UINT32, UString>::const_iterator iterator = m_NumberToStringMap.find(number);
  if (iterator != m_NumberToStringMap.end())
  {
    s = (*iterator).second;
    return true;
  }
  else
    return false;
};

bool CUINT32FromToUStringMap::GetNumber(const UString &s, UINT32 &number) const
{
  map<UString, UINT32>::const_iterator iterator =
      m_StringToNumberMap.find(s);
  if (iterator != m_StringToNumberMap.end())
  {
    number = (*iterator).second;
    return true;
  }
  else
    return false;
};

void COptionsCommon::AddType(UINT32 flag, wchar_t *name)
{
  COptionPair pair;
  pair.Flag = flag;
  pair.Name = name;
  m_Pairs.Add(pair);
}

void COptionsCommon::GetOptionNames(UINT32 flag, UStringVector &names, UINT32 &notUsedFlags) const
{
  notUsedFlags = flag;
  names.Clear();
  for (int i = 0; i < m_Pairs.Size(); i++)
    if ((m_Pairs[i].Flag & flag) != 0)
    {
      notUsedFlags &= (~m_Pairs[i].Flag);
      flag = notUsedFlags; // test it
      names.Add(m_Pairs[i].Name);
    }
};

bool COptionsCommon::GetFirstZeroFlagName(UString &name) const
{
  for (int i = 0; i < m_Pairs.Size(); i++)
    if (m_Pairs[i].Flag == 0)
    {
      name = m_Pairs[i].Name;
      return true;
    }
  return false;
}

bool COptionsCommon::GetOptionFlag(const UString &name, UINT32 &flag) const
{
  for (int i = 0; i < m_Pairs.Size(); i++)
    if (name.CompareNoCase(m_Pairs[i].Name) == 0)
    {
      flag = m_Pairs[i].Flag;
      return true;
    }
  return false;
};

void CDataInStream::SkipWords(UINT32 numWords)
{
  m_Pos += numWords;
}

void CDataInStream::SkipBytes(UINT32 numBytes)
{
  if ((numBytes & 1) != 0)
    throw 101853;
  SkipWords(numBytes / sizeof(UINT16));
}
  
void CDataInStream::SkipToAlignDWORD()
{
  if ((m_Pos & 1) != 0)
    GetWORD();
}

UINT16 CDataInStream::GetWORD()
{
  if (m_Pos >= m_Size)
    throw 101832; // CDecoderException();
  return m_Data[m_Pos++];
}

UINT32 CDataInStream::GetDWORD()
{
  UINT32 result = GetWORD();
  result |= (GetWORD() << 16);
  return result;
}

UString CDataInStream::GetString()
{
  UString result;
  wchar_t c;
  while ((c = GetWORD()) != 0)
    result += c;
  return result;
}


void CDataOutStream::PutWORD(UInt16 value)
{
  m_Data.Add(value);
}

void CDataOutStream::PutDWORD(UInt32 value)
{
  PutWORD(value & 0xFFFF);
  PutWORD(value >> 16);
}

void CDataOutStream::AlignToDWORD()
{
  if ((m_Data.Size() & 1) != 0)
    PutWORD(0);
}

void CDataOutStream::PutString(const UString &s)
{
  for (int i = 0; i < s.Length(); i++)
    m_Data.Add(s[i]);
  m_Data.Add(0);
}


void CTextInStream::Init(const UINT16 *data, UINT32 size)
{
  m_Text = L"";
  for (UINT32 i = 0; i < size; i++)
  {
    UINT16 c = data[i];
    if (c == 0)
      throw 1841;
    m_Text += c;
  }
  m_Pos = 0;
  m_Len = m_Text.Length();
}

static bool IsSeparatorChar(wchar_t c)
{
  return (c == kCharSpace || c == kCharTAB);
}

bool CTextInStream::WaitCharNotStepOver(wchar_t waitChar)
{
  while (m_Pos < m_Len)
  {
    wchar_t c = m_Text[m_Pos];
    if (c == waitChar)
      return true;
    if (c == kCharCR || c == kCharLF || IsSeparatorChar(c))
    {
      m_Pos++;
      continue;
    }
    return false;
  }
  return false;
}

bool CTextInStream::WaitChar(wchar_t c)
{
  if (WaitCharNotStepOver(c))
  {
    m_Pos++;
    return true;
  }
  return false;
}

bool CTextInStream::WaitComma()
{
  return WaitChar(kCharComma);
}

bool CTextInStream::WaitCommaNotStepOver()
{
  return WaitCharNotStepOver(kCharComma);
}

bool CTextInStream::WaitOrDelimiter()
{
  return WaitChar(kCharORDelimeter);
}

enum EMenuInputTextState
{
  kMenuStateEmpty,
  kMenuStateText,
  kMenuStateInQuote,
  kMenuStateInQuoteEnd
};

bool CTextInStream::GetSimpleString(UString &s)
{
  s = L"";
  bool textMode = false;
  while (m_Pos < m_Len)
  {
    wchar_t c = m_Text[m_Pos];
    if (!textMode)
    {
      if (c == kCharCR || c == kCharLF || IsSeparatorChar(c))
      {
        m_Pos++;
        continue;
      }
      textMode = true;
      continue;
    }
    else
    {
      if (c == kCharCR || c == kCharLF || IsSeparatorChar(c) ||
          c == kCharComma || c == kCharORDelimeter)
        return true;
      s += c;
      m_Pos++;
      continue;
    }
  }
  return (!s.IsEmpty());
}

bool CTextInStream::GetString(bool &inQuotes, UString &s)
{
  inQuotes = false;
  EMenuInputTextState state = kMenuStateEmpty;
  s = L"";
  while (m_Pos < m_Len)
  {
    wchar_t c = m_Text[m_Pos];
    switch(state)
    {
      case kMenuStateEmpty:
      {
        m_Pos++;
        if (c == kCharCR || c == kCharLF || IsSeparatorChar(c))
          continue;
        if (c == kCharQuote)
        {
          inQuotes = true;
          state = kMenuStateInQuote;
        }
        else
        {
          s += c;
          state = kMenuStateText;
        }
        continue;
      }
      case kMenuStateText:
      {
        m_Pos++;
        if (c == kCharCR || c == kCharLF || IsSeparatorChar(c))
          return true;
        s += c;
        continue;
      }
      case kMenuStateInQuote:
      {
        m_Pos++;
        if (c == kCharQuote)
          state = kMenuStateInQuoteEnd;
        else
          s += c;
        continue;
      }
      case kMenuStateInQuoteEnd:
      {
        if (c == kCharQuote)
          s += c;
        else
          return true;
        m_Pos++;
        state = kMenuStateInQuote;
        continue;
      }
    }
  }
  if (state == kMenuStateInQuote)
    throw 1921;
  return (!s.IsEmpty());
}

UString UnsignedToUString(UINT32 val)
{
  wchar_t temp[32];
  ConvertUInt32ToString(val, temp);
  return temp;
}

UString UnsignedToHexUString(UINT32 val)
{
  wchar_t temp[32];
  swprintf(temp, L"%X", val);
  return UString(L"0x") + temp;
}

CSysString DWORDToString(UINT32 val)
{
  TCHAR temp[32];
  ConvertUInt32ToString(val, temp);
  return temp;
}
