// ResDialog.cpp

#include "StdAfx.h"

#include "Common/StringConvert.h"

#include "CoderUtils.h"
#include "ResDialog.h"
#include "StringSubstitute.h"

namespace NWindows {
namespace NResource {
namespace NDialog {

static const wchar_t kCharCR = 13;
static const wchar_t kCharLF = 10;
static const wchar_t kCharTAB = 9;
static const wchar_t kCharQuote = L'\"';

static const wchar_t kCharSpace = L' ';
static const wchar_t kCharComma = L',';

static const wchar_t *kPropertyDelimiter = L", ";

static const wchar_t kCharORDelimeter = L'|';

static const wchar_t *kDialogString = L"DIALOG";
static const wchar_t *kDialogExString = L"DIALOGEX";

static const wchar_t *kCaptionString = L"CAPTION";
static const wchar_t *kClassString = L"CLASS";
static const wchar_t *kStyleString = L"STYLE";
static const wchar_t *kExStyleString = L"EXSTYLE";
static const wchar_t *kFontString = L"FONT";
static const wchar_t *kMenuString = L"MENU";

static const wchar_t *kDialogBeginString = L"{";
static const wchar_t *kDialogEndString = L"}";

static const UINT32 kMaxNumberValue = 0xFFFF;
static const UINT32 kMaxNumDigits = 5;

#pragma pack( push, PragmaDialogStructures)
#pragma pack( push, 1)

struct CDialogTemplate
{
  DWORD Style;
  DWORD ExtendedStyle;
  WORD  NumItems;
  short x;
  short y;
  short xSize;
  short ySize;

// wchar_t[] Menu
//   0x0000 - No Menu
//   0xFFFF + MenuId
//   null-terminated Unicode string  - Name;

// wchar_t[] ClassName
//   0x0000 - Predefined Class
//   0xFFFF + Class Id
//   null-terminated Unicode string  - Name;

// wchar_t[] Title
//   null-terminated Unicode string  - Name;

// if the style member specifies the DS_SETFONT
//      {
//        UINT16 FontSize;
//        wchar_t[] FontName; null-terminated Unicode string
//      }
};

// DLGITEMTEMPLATE structures must be aligned on DWORD boundaries

struct CDialogItemTemplate
{
  DWORD Style;
  DWORD ExtendedStyle;
  short x;
  short y;
  short xSize;
  short ySize;
  WORD  ID;
};

// Each DLGITEMTEMPLATE structure in the template must be aligned on a DWORD boundary

// The class and title arrays must be aligned on WORD boundaries.  ?(DWORD)
// The creation data array must be aligned on a WORD boundary.

// wchar_t[] Class
// if (Class[0] == 0xFFFF)
//    Class[1] means
//      0x0080 Button
//      0x0081 Edit
//      0x0082 Static
//      0x0083 List box
//      0x0084 Scroll bar
//      0x0085 Combo box
//  else Name Of Class

// wchar_t[] Title
// if (Title[0] == 0xFFFF)
//   Title[1] = specifies an ordinal value of a resource,
//               such as an icon, in an executable file
// else Initial Title;

// The creation data array begins at the next WORD boundary
// after the title array. This creation data can be of any
// size and format. If the first word of the creation data
// array is nonzero, it indicates the size, in bytes, of the
// creation data (including the size word). The control's window
// procedure must be able to interpret the data. When the system
// creates the control, it passes a pointer to this data in the
// lParam parameter of the WM_CREATE message that it sends to the control.


struct CDialogTemplateEx
{
  WORD      Version; // Must be 1
  WORD      Signature; // If signature is 0xFFFF, this is an extended dialog box template.
                       // In this case, the dlgVer member specifies the template version number.
  DWORD     helpID;
  DWORD     ExtendedStyle;
  DWORD     Style;
  WORD      NumItems;
  short     x;
  short     y;
  short     xSize;
  short     ySize;
  // wchar_t[] Menu
  // wchar_t[] ClassName
  // wchar_t[] Title
  // The following members exist only if the style member is
  // set to DS_SETFONT or DS_SHELLFONT.
  /*
  short     PointSize;
  short     Weight; // weight of the font in the range 0 through 1000
  short     Italic; // TRUE means Italic
  wchar_t[] FontName;
  */
} DLGTEMPLATEEX;

// Each DLGITEMTEMPLATEEX structure in the template must be aligned on a DWORD boundary

// The menu, windowClass, title, and typeface arrays must be aligned on WORD boundaries.

struct CDialogItemTemplateEx
{
  DWORD HelpID;
  DWORD ExtendedStyle;
  DWORD Style;
  short x;
  short y;
  short xSize;
  short ySize;
  WORD  ID;
  //// wchar_t[] Class
  // wchar_t[] Title
  // The creation data array begins at the next WORD boundary
};

// Each DLGITEMTEMPLATEEX structure must be aligned on a DWORD boundary.
// The variable-length windowClass and title arrays must be aligned on
// WORD boundaries. The creation data array, if any, must be aligned on a WORD boundary.


#pragma pack(pop)
#pragma pack(pop, PragmaDialogStructures)

//------------------------------------
// CDecoder

void CDecoder::GetProperties(CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_DIALOG);
}

// ----------------------
// CDialogStringSubstituter

static const CSubstituteInitPair kDialogSubstituteInitPairs[] =
{
  { L"\xA", L"\\n" },
  { L"\xD", L"\\r" },
  { L"\x8", L"\\a" },
  { L"\x9", L"\\t" },
  { L"\"", L"\"\"" },
  { L"\\", L"\\\\" }
};


class CDialogStringSubstituter
{
  CStringSubstituter m_Substituter;
public:
  CDialogStringSubstituter();
  UString ConvertBinToText(const UString &aBinString);
  UString ConvertTextToBin(const UString &aBinString);
};

CDialogStringSubstituter::CDialogStringSubstituter()
{
  m_Substituter.Init(kDialogSubstituteInitPairs,
    sizeof(kDialogSubstituteInitPairs) / sizeof(kDialogSubstituteInitPairs[0]));
}


UString GetQuotedString(const UString &s)
{
  UString textString;
  textString += kCharQuote;
  textString += s;
  textString += kCharQuote;
  return textString;
}

UString GetSpacedString(const UString &s)
{
  UString textString;
  textString += kCharSpace;
  textString += s;
  textString += kCharSpace;
  return textString;
}

UString CDialogStringSubstituter::ConvertBinToText(const UString &s)
{
  UString textString;
  textString += kCharQuote;
  textString += m_Substituter.Substitute(s, true);
  textString += kCharQuote;
  return textString;
}

UString CDialogStringSubstituter::ConvertTextToBin(const UString &s)
{
  return m_Substituter.Substitute(s, false);
}

static CDialogStringSubstituter g_DialogStringSubstituter;

// ----------------------------------------------------------

  
static void AddNewLine(UString &resultString)
{
  resultString += kCharCR;
  resultString += kCharLF;
}

class CStandardDialogItem;

typedef CRecordVector<CStandardDialogItem *> CStandardDialogItemVector;


//////////////////////////////////////

class CNumberOrUString
{
public:
  bool m_IsNumber;
  UINT16 m_Number;
  UString m_String;
  void SetAsNumber(UINT16 number);
  void SetAsString(const UInt16 *s);
};

/////////////////////////////////////////////////
// CNumberOrUString

void CNumberOrUString::SetAsNumber(UINT16 number)
{
  m_IsNumber = true;
  m_Number = number;
}

void CNumberOrUString::SetAsString(const UInt16 *s)
{
  m_IsNumber = false;
  m_String = (LPCWSTR)s;
}

namespace NClassesID
{
  enum
  {
    kButton = 0x0080,
    kEdit = 0x0081,
    kStatic = 0x0082,
    kListBox = 0x0083,
    kScrollBar = 0x0084,
    kComboBox = 0x0085
  };
}
namespace NClassesCaptions
{
  static const wchar_t *kButton = L"BUTTON";
  static const wchar_t *kEdit = L"EDITTEXT";
  static const wchar_t *kStatic = L"STATIC";
  static const wchar_t *kListBox = L"LISTBOX";
  static const wchar_t *kScrollBar = L"SCROLLBAR";
  static const wchar_t *kComboBox = L"COMBOBOX";
}

UString GetClassNameFromIDAndStyle(WORD id, DWORD /* style */, bool &printCaption)
{
  printCaption = true;
  switch(id)
  {
    case NClassesID::kButton: return NClassesCaptions::kButton;
    case NClassesID::kEdit: printCaption = false; return NClassesCaptions::kEdit;
    case NClassesID::kStatic: return NClassesCaptions::kStatic;
    case NClassesID::kListBox: return NClassesCaptions::kListBox;
    case NClassesID::kScrollBar: return NClassesCaptions::kScrollBar;
    case NClassesID::kComboBox: return NClassesCaptions::kComboBox;
  }
  throw 110425;
}

struct CControlInfoCommon
{
  CNumberOrUString m_Class;
  CNumberOrUString m_Title;
  CByteDynamicBuffer m_Data;
  UINT32 m_DataSize;
  
  void LoadFromBinaryCommon(CDataInStream &inStream);
};

struct CControlInfo: public CControlInfoCommon
{
  CDialogItemTemplate m_Info;
  
  void LoadFromBinary(CDataInStream &inStream);
  void SaveToText(UString &outStream, CIndentMaker &indentMaker) const ;
};

typedef CObjectVector<CControlInfo> CControlInfoVector;

struct CControlExInfo: public CControlInfoCommon
{
  CDialogItemTemplateEx m_Info;
  void LoadFromBinary(CDataInStream &inStream);
  void SaveToText(UString &outStream, CIndentMaker &indentMaker) const;
};

typedef CObjectVector<CControlExInfo> CControlExInfoVector;

// -----------------------------
// Load from Binary

static void ReadNumberOrString(CDataInStream &inStream, CNumberOrUString &m_NumberOrString)
{
  WORD word = inStream.GetWORD();
  if (word == 0xFFFF)
    m_NumberOrString.SetAsNumber(inStream.GetWORD());
  else
  {
    m_NumberOrString.SetAsString(inStream.GetData() - 1);
    inStream.SkipWords(m_NumberOrString.m_String.Length() - 1 + 1);
  }
}

void CControlInfoCommon::LoadFromBinaryCommon(CDataInStream &inStream)
{
  // inStream.SkipToAlignDWORD();
  ReadNumberOrString(inStream, m_Class);
  ReadNumberOrString(inStream, m_Title);

  m_DataSize = inStream.GetWORD();
  if (m_DataSize != 0)
  {
    if (m_DataSize < sizeof(UINT16))
      throw 101910;
    m_DataSize -= sizeof(UINT16);
    m_Data.EnsureCapacity(m_DataSize);
    memcpy((BYTE *)m_Data, inStream.GetData(), m_DataSize);
    inStream.SkipBytes(m_DataSize);
  }
}

void CControlInfo::LoadFromBinary(CDataInStream &inStream)
{
  inStream.SkipToAlignDWORD();
  memcpy(&m_Info, inStream.GetData(), sizeof(m_Info));
  inStream.SkipBytes(sizeof(m_Info));
  LoadFromBinaryCommon(inStream);
}

void CControlExInfo::LoadFromBinary(CDataInStream &inStream)
{
  inStream.SkipToAlignDWORD();
  memcpy(&m_Info, inStream.GetData(), sizeof(m_Info));
  inStream.SkipBytes(sizeof(m_Info));
  inStream.SkipToAlignDWORD(); // Test it
  LoadFromBinaryCommon(inStream);
}


void SaveCoordsToText(UString &outStream, int x, int y, int xSize, int ySize)
{
  outStream += UnsignedToUString(x);
  outStream += kPropertyDelimiter;
  outStream += UnsignedToUString(y);
  outStream += kPropertyDelimiter;
  outStream += UnsignedToUString(xSize);
  outStream += kPropertyDelimiter;
  outStream += UnsignedToUString(ySize);
}

UString GetClassNameOfControl(const CNumberOrUString &cls, DWORD style, bool &printCaption)
{
  printCaption = true;
  if (cls.m_IsNumber)
  {
    return GetClassNameFromIDAndStyle(cls.m_Number, style, printCaption);
  }
  else
  {
    return cls.m_String;
  }
}


void CControlInfo::SaveToText(UString &outStream, CIndentMaker & /* indentMaker */) const
{
  bool printCaption;
  outStream += GetClassNameOfControl(m_Class, m_Info.Style, printCaption);
  outStream += kCharSpace;
  if (printCaption)
  {
    if (m_Title.m_IsNumber)
      outStream += UnsignedToUString(m_Title.m_Number);
    else
      outStream += g_DialogStringSubstituter.ConvertBinToText(m_Title.m_String);
    outStream += kPropertyDelimiter;
  }
  outStream += UnsignedToUString(m_Info.ID);
  outStream += kPropertyDelimiter;
  SaveCoordsToText(outStream, m_Info.x, m_Info.y, m_Info.xSize, m_Info.ySize);
  if (m_Info.Style != 0 || m_Info.ExtendedStyle != 0)
  {
    outStream += kPropertyDelimiter;
    outStream += UnsignedToHexUString(m_Info.Style);
  }
  if (m_Info.ExtendedStyle != 0)
  {
    outStream += kPropertyDelimiter;
    outStream += UnsignedToHexUString(m_Info.ExtendedStyle);
  }
}

void CControlExInfo::SaveToText(UString &outStream, CIndentMaker & /* indentMaker */) const
{
  bool printCaption;
  outStream += GetClassNameOfControl(m_Class, m_Info.Style, printCaption);
  outStream += kCharSpace;
  if (printCaption)
  {
    if (m_Title.m_IsNumber)
      outStream += UnsignedToUString(m_Title.m_Number);
    else
      outStream += g_DialogStringSubstituter.ConvertBinToText(m_Title.m_String);
    outStream += kPropertyDelimiter;
  }
  outStream += UnsignedToUString(m_Info.ID);
  outStream += kPropertyDelimiter;
  SaveCoordsToText(outStream, m_Info.x, m_Info.y, m_Info.xSize, m_Info.ySize);
  if (m_Info.Style != 0 || m_Info.ExtendedStyle != 0)
  {
    outStream += kPropertyDelimiter;
    outStream += UnsignedToHexUString(m_Info.Style);
  }
  if (m_Info.ExtendedStyle != 0)
  {
    outStream += kPropertyDelimiter;
    outStream += UnsignedToHexUString(m_Info.ExtendedStyle);
  }
}

static UString NumberOrUString2String(const UString &name,
    const CNumberOrUString &a)
{
  UString outStream;
  if (a.m_IsNumber || !a.m_String.IsEmpty())
  {
    outStream += GetSpacedString(name);
    if (a.m_IsNumber)
      outStream += UnsignedToUString(a.m_Number);
    else
      outStream += GetQuotedString(a.m_String);
  }
  return outStream;
}


/////////////////////////////////
// CDialogInfo

class CDialogInfo
{
public:
  CDialogTemplate m_Info;

  CNumberOrUString m_Menu;
  CNumberOrUString m_ClassName;
  UString m_Title;

  bool m_FontSpecified;
  UINT16 m_FontSize;
  UString m_FontName;

  CControlInfoVector m_ControlInfoVector;

  void LoadFromBinary(CDataInStream &inStream);
  void SaveToText(UString &outStream, CIndentMaker &indentMaker);
};

void CDialogInfo::LoadFromBinary(CDataInStream &inStream)
{
  memcpy(&m_Info, inStream.GetData(), sizeof(m_Info));
  inStream.SkipBytes(sizeof(m_Info));
  ReadNumberOrString(inStream, m_Menu);
  ReadNumberOrString(inStream, m_ClassName);
  m_Title = inStream.GetString();
  m_FontSpecified = (m_Info.Style & DS_SETFONT) != 0;
  if (m_FontSpecified)
  {
    m_FontSize = inStream.GetWORD();;
    m_FontName = inStream.GetString();
  }
  for (int i = 0; i < m_Info.NumItems; i++)
  {
    m_ControlInfoVector.Add(CControlInfo());
    CControlInfo &controlInfo = m_ControlInfoVector.Back();
    controlInfo.LoadFromBinary(inStream);
  }
}

void CDialogInfo::SaveToText(UString &outStream, CIndentMaker &indentMaker)
{
  outStream += kDialogString;
  outStream += kCharSpace;
  SaveCoordsToText(outStream, m_Info.x, m_Info.y, m_Info.xSize, m_Info.ySize);

  if (!m_Title.IsEmpty())
  {
    outStream += GetSpacedString(kCaptionString);
    outStream += GetQuotedString(m_Title);
  }
  
  outStream += NumberOrUString2String(kClassString, m_ClassName);

  if (m_Info.Style != 0)
  {
    outStream += GetSpacedString(kStyleString);
    outStream += UnsignedToHexUString(m_Info.Style);
  }

  if (m_Info.ExtendedStyle != 0)
  {
    outStream += GetSpacedString(kExStyleString);
    outStream += UnsignedToHexUString(m_Info.ExtendedStyle);
  }

  outStream += NumberOrUString2String(kMenuString, m_Menu);

  if (m_FontSpecified)
  {
    outStream += GetSpacedString(kFontString);
    outStream += UnsignedToUString(m_FontSize);
    outStream += kPropertyDelimiter;
    outStream += GetQuotedString(m_FontName);
  }
  AddNewLine(outStream);

  outStream += kDialogBeginString;
  indentMaker.AddIndent();
  for (int i = 0; i < m_ControlInfoVector.Size(); i++)
  {
    AddNewLine(outStream);
    indentMaker.WriteIndent(outStream);
    const CControlInfo &controlInfo = m_ControlInfoVector[i];
    controlInfo.SaveToText(outStream, indentMaker);
  }
  indentMaker.DeleteIndent();
  AddNewLine(outStream);
  outStream += kDialogEndString;
  AddNewLine(outStream);

}


/////////////////////////////////
// CDialogExInfo

class CDialogExInfo
{
public:
  CDialogTemplateEx m_Info;

  CNumberOrUString m_Menu;
  CNumberOrUString m_ClassName;
  UString m_Title;

  bool m_FontSpecified;
  short m_FontPointSize;
  short m_FontWeight;
  short m_FontItalic;
  UString m_FontName;
  
  CControlExInfoVector m_ControlInfoVector;

  void LoadFromBinary(CDataInStream &inStream);
  void SaveToText(UString &outStream, CIndentMaker &indentMaker);
};

void CDialogExInfo::LoadFromBinary(CDataInStream &inStream)
{
  memcpy(&m_Info, inStream.GetData(), sizeof(m_Info));
  inStream.SkipBytes(sizeof(m_Info));
  ReadNumberOrString(inStream, m_Menu);
  ReadNumberOrString(inStream, m_ClassName);
  m_Title = inStream.GetString();
  m_FontSpecified = (m_Info.Style & DS_SETFONT) != 0;
  if (m_FontSpecified)
  {
    m_FontPointSize = inStream.GetWORD();;
    m_FontWeight = inStream.GetWORD();;
    m_FontItalic = inStream.GetWORD();;
    m_FontName = inStream.GetString();
  }
  for (int i = 0; i < m_Info.NumItems; i++)
  {
    m_ControlInfoVector.Add(CControlExInfo());
    CControlExInfo &controlInfo = m_ControlInfoVector.Back();
    controlInfo.LoadFromBinary(inStream);
  }
}

void CDialogExInfo::SaveToText(UString &outStream, CIndentMaker &indentMaker)
{
  outStream += kDialogExString;
  outStream += kCharSpace;
  SaveCoordsToText(outStream, m_Info.x, m_Info.y, m_Info.xSize, m_Info.ySize);

  if (!m_Title.IsEmpty())
  {
    outStream += GetSpacedString(kCaptionString);
    outStream += GetQuotedString(m_Title);
  }
  
  outStream += NumberOrUString2String(kClassString, m_ClassName);

  if (m_Info.Style != 0)
  {
    outStream += GetSpacedString(kStyleString);
    outStream += UnsignedToHexUString(m_Info.Style);
  }

  if (m_Info.ExtendedStyle != 0)
  {
    outStream += GetSpacedString(kExStyleString);
    outStream += UnsignedToHexUString(m_Info.ExtendedStyle);
  }

  outStream += NumberOrUString2String(kMenuString, m_Menu);

  if (m_FontSpecified)
  {
    outStream += GetSpacedString(kFontString);
    outStream += UnsignedToUString(m_FontPointSize);
    outStream += kPropertyDelimiter;
    outStream += GetQuotedString(m_FontName);
    outStream += kPropertyDelimiter;
    outStream += UnsignedToUString(m_FontWeight);
    outStream += kPropertyDelimiter;
    outStream += UnsignedToUString(m_FontItalic);
  }
  AddNewLine(outStream);

  outStream += kDialogBeginString;
  indentMaker.AddIndent();
  for (int i = 0; i < m_ControlInfoVector.Size(); i++)
  {
    AddNewLine(outStream);
    indentMaker.WriteIndent(outStream);
    const CControlExInfo &controlInfo = m_ControlInfoVector[i];
    controlInfo.SaveToText(outStream, indentMaker);
  }
  indentMaker.DeleteIndent();
  AddNewLine(outStream);
  outStream += kDialogEndString;
  AddNewLine(outStream);

}

//////////////////////////////////////////////////
void CDecoder::ReadItems(const CItemNameID &resourceName,
    const CDataInfo &resourceData,
    CSubResourcePropertiesVector &structure) const
{
  structure.Add(CSubResourceProperties());
  CSubResourceProperties &resourceProps = structure.Back();

  resourceProps.Name = resourceName;
  resourceProps.DefaultExtension = TEXT("txt");

  UINT32 resourceSize = resourceData.Size / 2;
  UINT16 *pData = (UINT16 *)((BYTE *)resourceData.Data);
  CDataInStream inStream(pData, resourceSize);
  UINT16 version = inStream.PeekWORD(0);
  UINT16 signature = inStream.PeekWORD(1);
  CIndentMaker indentMaker;
  UString outStream;
  if (signature == 0xFFFF)
  {
    if (version != 1)
      throw "Unknown version";
    CDialogExInfo dialogInfo;
    dialogInfo.LoadFromBinary(inStream);
    dialogInfo.SaveToText(outStream, indentMaker);
    resourceProps.Description = GetSystemString(dialogInfo.m_Title);
  }
  else
  {
    CDialogInfo dialogInfo;
    dialogInfo.LoadFromBinary(inStream);
    dialogInfo.SaveToText(outStream, indentMaker);
    resourceProps.Description = GetSystemString(dialogInfo.m_Title);
  }

  resourceProps.Size = outStream.Length() * sizeof(wchar_t);
  resourceProps.Data.EnsureCapacity(resourceProps.Size);
  memcpy(resourceProps.Data, (LPCWSTR)outStream, resourceProps.Size);
}

//------------------------------------
// CEncoder

void CEncoder::GetProperties(CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_DIALOG);
}

void CEncoder::WriteItems(const CSubResourcePropertiesVector &structure,
    CDataBlock & /* resourceData */) const
{
  if (structure.Size() != 1)
    throw 2053;
  const CSubResourceProperties &subResourceProps = structure[0];

  CTextInStream inStream;
  if ((subResourceProps.Size & 1) != 0)
    throw 242243;
  inStream.Init((const UINT16 *)((const BYTE *)subResourceProps.Data),
    subResourceProps.Size / sizeof(wchar_t));
  UString s;
  if (!inStream.GetSimpleString(s))
    throw 20128;
  CDataOutStream outStream;
  if (s.CompareNoCase(kDialogString) == 0)
  {
    throw "Update Dialog is not supported";
  }
  else if (s.CompareNoCase(kDialogExString) == 0)
  {
    throw "Update DialogEx is not supported";
  }
  else
    throw 20131;
  /*
  resourceData.Size = outStream.m_Data.Size() * sizeof(UInt16);
  resourceData.Data.EnsureCapacity(resourceData.Size);

  memcpy(resourceData.Data, &outStream.m_Data.Front(), resourceData.Size);
  */
}

}}}
