// ResIcon.cpp

#include "StdAfx.h"

#include "Common/IntToString.h"

#include "CoderUtils.h"
#include "IconCursor.h"
#include "ResIcon.h"

namespace NWindows {
namespace NResource {

using namespace NIconCursor;

static const TCHAR kDelimiterChar = TEXT('*');

static UInt32 GetImageSize(UInt32 xSize, UInt32 ySize, UInt32 bitCount)
{
  return ((xSize * bitCount + 7) / 8 + 3) / 4 * 4 * ySize;
}

namespace NIcon {

void CDecoder::GetProperties(CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_ICON);
}

static const Byte kPngSig[] = { 0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A };

void CDecoder::ReadItems(const CItemNameID &resourceName,
    const CDataInfo &resourceData, CSubResourcePropertiesVector &structure) const
{
  structure.Add(CSubResourceProperties());
  CSubResourceProperties &resourceProps = structure.Back();

  UInt32 resourceSize = resourceData.Size;

  resourceProps.Name = resourceName;
  const Byte *data = (const Byte *)resourceData.Data;
  if (resourceSize < 8)
    throw CDecoderException();
  CByteDynamicBuffer &buf = resourceProps.Data;
  if (memcmp(data, kPngSig, 8) == 0)
  {
    resourceProps.DefaultExtension = TEXT("png");
    resourceProps.Size = resourceSize;
    buf.EnsureCapacity(resourceSize);
    memcpy(buf, data, resourceSize);
    return;
  }

  resourceProps.DefaultExtension = TEXT("ico");

  if (resourceSize < sizeof(BITMAPINFOHEADER))
    throw CDecoderException();

  BITMAPINFOHEADER &info = *((BITMAPINFOHEADER *)(data));
  if (info.biSize != sizeof(info))
    throw CDecoderException();
  
  UInt32 numBitCount = info.biBitCount;
  if (numBitCount != 1 &&
      numBitCount != 4 &&
      numBitCount != 8 &&
      numBitCount != 24 &&
      numBitCount != 32)
    throw CDecoderException();

  UInt32 xSize = info.biWidth;
  if ((info.biHeight & 1) != 0)
    throw CDecoderException();
  UInt32 ySize = info.biHeight / 2;
  if (xSize > 0x100 || ySize > 0x100)
    throw CDecoderException();

  UInt32 imageSize;
  // imageSize is not correct if AND mask array contains zeros
  // in this case it is equal image1Size

  // UInt32 imageSize = info.biSizeImage;
  // if (imageSize == 0)
  // {
    if (info.biCompression != BI_RGB)
      throw CDecoderException();
    UInt32 image1Size = GetImageSize(xSize, ySize, numBitCount);
    UInt32 image2Size = GetImageSize(xSize, ySize, 1);
    imageSize = image1Size + image2Size;
  // }
  UInt32 numColors = 0;
  if (numBitCount < 16)
    numColors = 1 << numBitCount;

  CHeader iconHeader;
  iconHeader.ResType = RES_ICON;
  iconHeader.ResCount = 1;
  iconHeader.Reserved = 0;

  CResDirInFile iconItemInfo;
  iconItemInfo.Icon.Width = (Byte)xSize;
  iconItemInfo.Icon.Height = (Byte)ySize;
  iconItemInfo.Icon.ColorCount = (Byte)numColors;
  iconItemInfo.Icon.Reserved = 0;
  iconItemInfo.Reserved1 = 0;
  iconItemInfo.Reserved2 = 0;
  UInt32 numQuadsBytes = numColors * 4;
  iconItemInfo.BytesInRes = sizeof(BITMAPINFOHEADER) + numQuadsBytes + imageSize;
  UInt32 headersSize = sizeof(iconHeader) + sizeof(iconItemInfo);
  iconItemInfo.ImageOffset = headersSize;

  if (iconItemInfo.BytesInRes != resourceSize)
    throw CDecoderException();

  resourceProps.Description = DWORDToString(xSize) +
      kDelimiterChar + DWORDToString(ySize) +
      kDelimiterChar + DWORDToString(numBitCount);
  
  resourceProps.Size = headersSize + resourceSize;
  buf.EnsureCapacity(resourceProps.Size);
  memcpy(buf, &iconHeader, sizeof(iconHeader));
  memcpy(buf + sizeof(iconHeader), &iconItemInfo, sizeof(iconItemInfo));
  memcpy(buf + headersSize, resourceData.Data, resourceSize);
}

void CEncoder::GetProperties(CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_ICON);
}

static const char *kIsNotIconMessage = "Bad ICON data";
static const char *kMoreThanOneIcons = "Icon data contain more than one image";

void CEncoder::WriteItems(const CSubResourcePropertiesVector &structure,
    CDataBlock &resourceData) const
{
  if (structure.Size() != 1)
    throw 2053;
  const CSubResourceProperties &subResourceProps = structure[0];
  UInt32 headersSize = sizeof(CHeader) + sizeof(CResDirInFile);
  if (subResourceProps.Size < headersSize)
    throw kIsNotIconMessage;
  const CHeader &iconHeader = *((const CHeader *)(const Byte *)subResourceProps.Data);
  if (iconHeader.Reserved != 0 || iconHeader.ResType != RES_ICON || iconHeader.ResCount == 0)
    throw kIsNotIconMessage;
  if (iconHeader.ResCount > 1)
    throw kMoreThanOneIcons;
  const CResDirInFile &iconItemInfo =
      *((const CResDirInFile *)(((const Byte *)subResourceProps.Data) +
        sizeof(CHeader)));
  UInt32 size = subResourceProps.Size - headersSize;
  if (iconItemInfo.Icon.Reserved != 0 || iconItemInfo.ImageOffset != headersSize ||
      iconItemInfo.BytesInRes != size)
    throw kIsNotIconMessage;
  resourceData.Data.EnsureCapacity(size);
  resourceData.Size = size;
  memcpy(resourceData.Data, subResourceProps.Data + headersSize, size);
}

}

// -----------------------
// Cursor

namespace NCursor {

void CDecoder::GetProperties(CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_CURSOR);
}

void CDecoder::ReadItems(const CItemNameID &resourceName,
    const CDataInfo &resourceData,
    CSubResourcePropertiesVector &structure) const
{
  structure.Add(CSubResourceProperties());
  CSubResourceProperties &resourceProps = structure.Back();

  UInt32 resourceSize = resourceData.Size;

  resourceProps.Name = resourceName;
  resourceProps.DefaultExtension = TEXT("cur");
  const Byte *data = (const Byte *)resourceData.Data;

  if (resourceSize < sizeof(CCursorSpots) + sizeof(BITMAPINFOHEADER))
    throw CDecoderException();

  const CCursorSpots &spots = *((const CCursorSpots *)(data));
  const BITMAPINFOHEADER &info = *((const BITMAPINFOHEADER *)(data +
      sizeof(CCursorSpots)));
  if (info.biSize != sizeof(info))
    throw CDecoderException();
  
  UInt32 numBitCount = info.biBitCount;
  if (numBitCount != 1 && numBitCount != 4 && numBitCount != 8)
    throw CDecoderException();

  UInt32 xSize = info.biWidth;
  if ((info.biHeight & 1) != 0)
    throw CDecoderException();
  UInt32 ySize = info.biHeight / 2;
  if (xSize >= 0x100 || ySize >= 0x100)
    throw CDecoderException();

  UInt32 imageSize;
  // imageSize is not correct if AND mask array contains zeros
  // in this case it is equal image1Size

  // UInt32 imageSize = info.biSizeImage;
  // if (imageSize == 0)
  // {
    if (info.biCompression != BI_RGB)
      throw CDecoderException();
    UInt32 image1Size = GetImageSize(xSize, ySize, numBitCount);
    UInt32 image2Size = GetImageSize(xSize, ySize, 1);
    imageSize = image1Size + image2Size;
  // }
  UInt32 numColors = 1 << numBitCount;

  CHeader cursorHeader;
  cursorHeader.ResType = RES_CURSOR;
  cursorHeader.ResCount = 1;
  cursorHeader.Reserved = 0;

  CResDirInFile cursorItemInfo;
  cursorItemInfo.Cursor.Width = (Byte)xSize;
  cursorItemInfo.Cursor.Height = (Byte)ySize;
  cursorItemInfo.Cursor.Reserved1 = 0;
  cursorItemInfo.Cursor.Reserved2 = 0;
  cursorItemInfo.XSpot = spots.XSpot;
  cursorItemInfo.YSpot = spots.YSpot;
  UInt32 numQuadsBytes = numColors * 4;
  cursorItemInfo.BytesInRes = sizeof(BITMAPINFOHEADER) + numQuadsBytes + imageSize;

  UInt32 headersSize = sizeof(cursorHeader) + sizeof(cursorItemInfo);
  cursorItemInfo.ImageOffset = headersSize;

  if (cursorItemInfo.BytesInRes != resourceSize - sizeof(CCursorSpots))
    throw CDecoderException();

  resourceProps.Description = DWORDToString(xSize)
      + kDelimiterChar + DWORDToString(ySize) +
      kDelimiterChar + DWORDToString(numBitCount);
  
  CByteDynamicBuffer &buf = resourceProps.Data;

  resourceProps.Size = headersSize + cursorItemInfo.BytesInRes;
  buf.SetCapacity(resourceProps.Size);
  memcpy(buf, &cursorHeader, sizeof(cursorHeader));
  memcpy(buf + sizeof(cursorHeader), &cursorItemInfo, sizeof(cursorItemInfo));
  memcpy(buf + headersSize, data + sizeof(CCursorSpots),
      cursorItemInfo.BytesInRes);
}

void CEncoder::GetProperties(CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_CURSOR);
}

static const char *kIsNotCursorMessage = "Bad CURSOR data";
static const char *kMoreThanOneCursors = "Cursor data contain more than one image";

void CEncoder::WriteItems(const CSubResourcePropertiesVector &structure,
    CDataBlock &resourceData) const
{
  if (structure.Size() != 1)
    throw 2053;
  const CSubResourceProperties &subResourceProps = structure[0];
  UInt32 headersSize = sizeof(CHeader) + sizeof(CResDirInFile);
  if (subResourceProps.Size < headersSize)
    throw kIsNotCursorMessage;
  const CHeader &cursorHeader = *((const CHeader *)(const Byte *)subResourceProps.Data);
  if (cursorHeader.Reserved != 0 || cursorHeader.ResType != RES_CURSOR ||
      cursorHeader.ResCount == 0)
    throw kIsNotCursorMessage;
  if (cursorHeader.ResCount > 1)
    throw kMoreThanOneCursors;
  const CResDirInFile &cursorItemInfo =
      *((const CResDirInFile *)(((const Byte *)subResourceProps.Data) +
        sizeof(CHeader)));
  UInt32 size = subResourceProps.Size - headersSize;
  if (cursorItemInfo.ImageOffset != headersSize ||
        cursorItemInfo.BytesInRes != size)
    throw kIsNotCursorMessage;

  CCursorSpots spots;
  spots.XSpot = cursorItemInfo.XSpot;
  spots.YSpot = cursorItemInfo.YSpot;

  resourceData.Size = size + sizeof(CCursorSpots);
  resourceData.Data.EnsureCapacity(resourceData.Size);
  memcpy(resourceData.Data, &spots, sizeof(CCursorSpots));
  memcpy(resourceData.Data + sizeof(CCursorSpots),  subResourceProps.Data + headersSize, size);
}

}

}}
