// ResMenu.cpp

#include "StdAfx.h"

#include "CoderUtils.h"
#include "ResMenu.h"
#include "StringSubstitute.h"

namespace NWindows {
namespace NResource {
namespace NMenu {

static const wchar_t kCharCR = 13;
static const wchar_t kCharLF = 10;
static const wchar_t kCharTAB = 9;
static const wchar_t kCharQuote = L'\"';

static const wchar_t kCharSpace = L' ';
static const wchar_t kCharComma = L',';

static const wchar_t kCharORDelimeter = L'|';

static const wchar_t *kStandardMenuString = L"MENU";
static const wchar_t *kMenuExString = L"MENUEX";

static const wchar_t *kMenuPopUpString = L"POPUP";
static const wchar_t *kMenuItemString = L"MENUITEM";
static const wchar_t *kMenuBeginString = L"{";
static const wchar_t *kMenuEndString = L"}";
static const wchar_t *kMenuSeparatorString = L"SEPARATOR";

static const wchar_t *kMenuExSeparatorString = L"MFT_SEPARATOR";

static const UINT32 kMenuFlagEnd = 0x80;
static const UINT32 kMenuFlagPopUp = 0x10;

static const UINT32 kMenuExFlagEnd = 0x80;
static const UINT32 kMenuExFlagPopUp = 0x01;

static const UINT32 kMaxNumberValue = 0xFFFF;
static const UINT32 kMaxNumDigits = 5;

static bool UStringToWORD(const UString &value, UINT16 &result)
{
  if (value.Length() > kMaxNumDigits)
    return false;
  for (int i = 0; i < value.Length() ; i++)
    if (!iswdigit(value[i]))
      return false;
  wchar_t *endptr;
  UINT32 dwordResult = wcstoul(value, &endptr, 10);
  result = UINT16(dwordResult);
  return (dwordResult <= kMaxNumberValue);
}

static bool UStringToDWORD(const UString &value, UINT32 &result)
{
  if (value.Length() > 12) // test it !! It was kMaxNumDigits
    return false;
  for (int i = 0; i < value.Length() ; i++)
    if (!iswdigit(value[i]))
      return false;
  wchar_t *endptr;
  result = wcstoul(value, &endptr, 10);
  return true;
}


//------------------------------------
// CDecoder

void CDecoder::GetProperties(CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_MENU);
}


class CResourceMenuOptions: public COptionsCommon
{
public:
  CResourceMenuOptions();
};

CResourceMenuOptions::CResourceMenuOptions()
{
  AddType(MF_CHECKED, L"CHECKED");
  AddType(MF_GRAYED, L"GRAYED");
  AddType(MF_HELP, L"HELP");
  AddType(MF_DISABLED, L"INACTIVE");
  AddType(MF_MENUBARBREAK, L"MENUBARBREAK");
  AddType(MF_MENUBREAK, L"MENUBREAK");
}

CResourceMenuOptions g_MenuOptions;

class CResourceMenuExTypeFlags: public COptionsCommon
{
public:
  CResourceMenuExTypeFlags();
};

CResourceMenuExTypeFlags::CResourceMenuExTypeFlags()
{
  AddType(MFT_BITMAP, L"MFT_BITMAP");
  AddType(MFT_MENUBARBREAK, L"MFT_MENUBARBREAK");
  AddType(MFT_MENUBREAK, L"MFT_MENUBREAK");
  AddType(MFT_OWNERDRAW, L"MFT_OWNERDRAW");
  AddType(MFT_RADIOCHECK, L"MFT_RADIOCHECK");
  AddType(MFT_RIGHTJUSTIFY, L"MFT_RIGHTJUSTIFY");
  AddType(MFT_RIGHTORDER, L"MFT_RIGHTORDER");
  AddType(MFT_SEPARATOR, L"MFT_SEPARATOR");
  AddType(MFT_STRING, L"MFT_STRING");
}

CResourceMenuExTypeFlags g_MenuExTypeFlags;

class CResourceMenuExStateFlags: public COptionsCommon
{
public:
  CResourceMenuExStateFlags();
};

CResourceMenuExStateFlags::CResourceMenuExStateFlags()
{
  AddType(MFS_CHECKED, L"MFS_CHECKED");
  AddType(MFS_DEFAULT, L"MFS_DEFAULT");
  AddType(MFS_GRAYED, L"MFS_GRAYED");
  AddType(MFS_DISABLED, L"MFS_DISABLED");
  AddType(MFS_ENABLED, L"MFS_ENABLED");
  AddType(MFS_HILITE, L"MFS_HILITE");
  AddType(MFS_UNCHECKED, L"MFS_UNCHECKED");
  AddType(MFS_UNHILITE, L"MFS_UNHILITE");
}

CResourceMenuExStateFlags g_MenuExStateFlags;


// ----------------------------------------
// CDataInStream



static bool TestFlag(UINT32 flag, UINT32 value)
{
  return ((value & flag) != 0);
}


// ----------------------
// CMenuStringSubstituter

static const CSubstituteInitPair kMenuSubstituteInitPairs[] =
{
  { L"\xA", L"\\n" },
  { L"\xD", L"\\r" },
  { L"\x8", L"\\a" },
  { L"\x9", L"\\t" },
  { L"\"", L"\"\"" },
  { L"\\", L"\\\\" }
};


class CMenuStringSubstituter
{
  CStringSubstituter m_Substituter;
public:
  CMenuStringSubstituter();
  UString ConvertBinToText(const UString &aBinString);
  UString ConvertTextToBin(const UString &aBinString);
};

CMenuStringSubstituter::CMenuStringSubstituter()
{
  m_Substituter.Init(kMenuSubstituteInitPairs,
    sizeof(kMenuSubstituteInitPairs) / sizeof(kMenuSubstituteInitPairs[0]));
}

UString CMenuStringSubstituter::ConvertBinToText(const UString &s)
{
  UString textString;
  textString += kCharQuote;
  textString += m_Substituter.Substitute(s, true);
  textString += kCharQuote;
  return textString;
}

UString CMenuStringSubstituter::ConvertTextToBin(const UString &s)
{
  return m_Substituter.Substitute(s, false);
}

static CMenuStringSubstituter g_MenuStringSubstituter;

static void AddNewLine(UString &resultString)
{
  resultString += kCharCR;
  resultString += kCharLF;
}

class CStandardMenuItem;

typedef CRecordVector<CStandardMenuItem *> CStandardMenuItemVector;

class CStandardMenuItem
{
public:
  UINT16 m_Flags;
  CStandardMenuItemVector m_SubMenus;
  UString m_Text;
  UINT16 m_ID;

  CStandardMenuItem(): m_Flags(0), m_ID(0) {};

  ~CStandardMenuItem();
  bool IsPopup() const { return TestFlag(kMenuFlagPopUp, m_Flags); };
  bool IsLast() const { return TestFlag(kMenuFlagEnd, m_Flags); };
  bool IsSeparator() const { return (m_ID == 0 && m_Text.Length() == 0 && !IsPopup()); };

  void LoadFromBinary(CDataInStream &inStream);
  static void LoadListFromBinary(CDataInStream &inStream,
      CStandardMenuItemVector &subMenus);
  
  void SaveToBinary(CDataOutStream &outStream);
  static void SaveListToBinary(CDataOutStream &outStream,
      CStandardMenuItemVector &subMenus);

  static void LoadListFromText(CTextInStream &inStream,
      CStandardMenuItemVector &subMenus);

  void SaveToText(UString &outStream, CIndentMaker &indentMaker);
  static void SaveListToText(UString &outStream, CIndentMaker &indentMaker,
      CStandardMenuItemVector &subMenus);
};


CStandardMenuItem::~CStandardMenuItem()
{
  for (int i = 0; i < m_SubMenus.Size(); i++)
    delete m_SubMenus[i];
}

// -----------------------------
// Load from Binary

void CStandardMenuItem::LoadFromBinary(CDataInStream &inStream)
{
  m_Flags = inStream.GetWORD();
  if (!IsPopup())
    m_ID = inStream.GetWORD();
  m_Text = inStream.GetString();
  if (IsPopup())
    LoadListFromBinary(inStream, m_SubMenus);
}

void CStandardMenuItem::LoadListFromBinary(CDataInStream &inStream,
    CStandardMenuItemVector &subMenus)
{
  for (;;)
  {
    subMenus.Add(new CStandardMenuItem());
    CStandardMenuItem *item = subMenus.Back();
    item->LoadFromBinary(inStream);
    if (item->IsLast())
      break;
  }
}

// -----------------------------------------
// Save to Binary

void CStandardMenuItem::SaveToBinary(CDataOutStream &outStream)
{
  outStream.PutWORD(m_Flags);
  if (!IsPopup())
    outStream.PutWORD(m_ID);
  outStream.PutString(m_Text);
  if (IsPopup())
    SaveListToBinary(outStream, m_SubMenus);
}

void CStandardMenuItem::SaveListToBinary(CDataOutStream &outStream,
    CStandardMenuItemVector &subMenus)
{
  for (int i = 0; i < subMenus.Size(); i++)
    subMenus[i]->SaveToBinary(outStream);
}

// -----------------------------
// Load from Text

void CStandardMenuItem::LoadListFromText(CTextInStream &inStream,
    CStandardMenuItemVector &subMenus)
{
  UString s;
  if (!inStream.GetSimpleString(s))
    throw 11711;
  if (s.CompareNoCase(kMenuBeginString) != 0)
    throw 11712;
  for (;;)
  {
    if (!inStream.GetSimpleString(s))
      throw 11704;
    if (s.CompareNoCase(kMenuEndString) == 0)
    {
      if (subMenus.Size() == 0)
        throw 11715;
      subMenus.Back()->m_Flags |= kMenuFlagEnd;
      return;
    }
    subMenus.Add(new CStandardMenuItem());
    CStandardMenuItem *item = subMenus.Back();
    
    item->m_Flags = 0;
    if (s.CompareNoCase(kMenuPopUpString) == 0)
      item->m_Flags |= kMenuFlagPopUp;
    else if (s.CompareNoCase(kMenuItemString) != 0)
      throw 242305;
    
    UString caption;
    bool quoted;
    if (!inStream.GetString(quoted, caption))
      throw 242039;
    
    if (!quoted)
    {
      if (item->IsPopup())
        throw 2423066;
      if (caption.CompareNoCase(kMenuSeparatorString) != 0)
        throw 242307;
      item->m_ID = 0;
      continue;
    }
    if (!item->IsPopup())
    {
      if (!inStream.WaitComma())
        throw 242235;
      if (!inStream.GetSimpleString(s))
        throw 242040;
      if (!UStringToWORD(s, item->m_ID))
        throw 242041;
    }
    item->m_Text = g_MenuStringSubstituter.ConvertTextToBin(caption);
    for (;;)
    {
      if (!inStream.WaitComma())
        break;
      if (!inStream.GetSimpleString(s))
        throw 242236;
      UINT32 flag;
      if (!g_MenuOptions.GetOptionFlag(s, flag))
        throw 242143;
      item->m_Flags |= UINT16(flag);
    }
    if (item->IsPopup())
      LoadListFromText(inStream, item->m_SubMenus);
  }
}

// ----------------------------
// Save to Text


void CStandardMenuItem::SaveToText(UString &outStream, CIndentMaker &indentMaker)
{
  indentMaker.WriteIndent(outStream);
  if (IsPopup())
    outStream += kMenuPopUpString;
  else
    outStream += kMenuItemString;
  outStream += kCharSpace;
  if (IsSeparator())
    outStream += kMenuSeparatorString;
  else
  {
    outStream += g_MenuStringSubstituter.ConvertBinToText(m_Text);
    if (!IsPopup())
    {
      outStream += kCharComma;
      outStream += kCharSpace;
      outStream += UnsignedToUString(m_ID);
    }
    UStringVector optionNames;
    UINT32 notUsedFlags;
    g_MenuOptions.GetOptionNames(m_Flags, optionNames, notUsedFlags);
    notUsedFlags &= (~kMenuFlagEnd);
    notUsedFlags &= (~kMenuFlagPopUp);
    if (notUsedFlags != 0)
      throw CDecoderException();
    for (int i = 0; i < optionNames.Size(); i++)
    {
      outStream += kCharComma;
      outStream += kCharSpace;
      outStream += optionNames[i];
    }
    
    // Test not used flags
    
    if (IsPopup())
    {
      AddNewLine(outStream);
      SaveListToText(outStream, indentMaker, m_SubMenus);
    }
  }
  AddNewLine(outStream);
}

void CStandardMenuItem::SaveListToText(UString &outStream, CIndentMaker &indentMaker,
    CStandardMenuItemVector &subMenus)
{
  indentMaker.WriteIndent(outStream);
  outStream += kMenuBeginString;
  AddNewLine(outStream);
  indentMaker.AddIndent();

  for (int i = 0; i < subMenus.Size(); i++)
    subMenus[i]->SaveToText(outStream, indentMaker);

  indentMaker.DeleteIndent();
  indentMaker.WriteIndent(outStream);
  outStream += kMenuEndString;
}


class CMenuExItem;

typedef CRecordVector<CMenuExItem *> CMenuExItemList;

class CMenuExItem
{
public:
  CMenuExItemList m_SubMenus;
  UINT32 m_Type;
  UINT32 m_State;
  UINT m_ID;
  UINT16 m_ResInfo;
  UString m_Text;
  UINT32 m_HelpID;

  // CMenuExItem(): m_Flags(0), m_ID(0) {};

  ~CMenuExItem();

  bool IsPopup() const { return TestFlag(kMenuExFlagPopUp, m_ResInfo); };
  bool IsLast() const { return TestFlag(kMenuExFlagEnd, m_ResInfo); };
  bool IsSeparator() const { return (/*m_ID == 0 && */m_Text.Length() == 0
      && !IsPopup() && m_Type == MFT_SEPARATOR); };  // test it

  void LoadFromBinary(CDataInStream &inStream);
  static void LoadListFromBinary(CDataInStream &inStream,
      CMenuExItemList &subMenus);
  
  void SaveToBinary(CDataOutStream &outStream);
  static void SaveListToBinary(CDataOutStream &outStream,
      CMenuExItemList &subMenus);

  static void LoadListFromText(CTextInStream &inStream,
      CMenuExItemList &subMenus);

  void SaveToText(UString &outStream, CIndentMaker &indentMaker);
  static void SaveListToText(UString &outStream, CIndentMaker &indentMaker,
      CMenuExItemList &subMenus);
};


CMenuExItem::~CMenuExItem()
{
  for (int i = 0; i < m_SubMenus.Size(); i++)
    delete m_SubMenus[i];
}

// -----------------------------
// Load from Binary

void CMenuExItem::LoadFromBinary(CDataInStream &inStream)
{
  inStream.SkipToAlignDWORD();
  m_Type = inStream.GetDWORD();
  m_State = inStream.GetDWORD();
  m_ID = inStream.GetDWORD();
  m_ResInfo = inStream.GetWORD();
  m_Text = inStream.GetString();
  if (IsPopup())
  {
    inStream.SkipToAlignDWORD();
    m_HelpID = inStream.GetDWORD();
    LoadListFromBinary(inStream, m_SubMenus);
  }
}

void CMenuExItem::LoadListFromBinary(CDataInStream &inStream,
    CMenuExItemList &subMenus)
{
  for (;;)
  {
    subMenus.Add(new CMenuExItem());
    CMenuExItem *item = subMenus.Back();
    item->LoadFromBinary(inStream);
    if (item->IsLast())
      break;
  }
}

// -----------------------------------------
// Save to Binary

void CMenuExItem::SaveToBinary(CDataOutStream &outStream)
{
  outStream.AlignToDWORD();
  outStream.PutDWORD(m_Type);
  outStream.PutDWORD(m_State);
  outStream.PutDWORD(m_ID);
  outStream.PutWORD(m_ResInfo);
  outStream.PutString(m_Text);
  if (IsPopup())
  {
    outStream.AlignToDWORD();
    outStream.PutDWORD(m_HelpID);
    SaveListToBinary(outStream, m_SubMenus);
  }
}

void CMenuExItem::SaveListToBinary(CDataOutStream &outStream,
    CMenuExItemList &subMenus)
{
  for (int i = 0; i < subMenus.Size(); i++)
    subMenus[i]->SaveToBinary(outStream);
}

// -----------------------------
// Load from Text

void CMenuExItem::LoadListFromText(CTextInStream &inStream,
    CMenuExItemList &subMenus)
{
  UString s;
  if (!inStream.GetSimpleString(s))
    throw 11711;
  if (s.CompareNoCase(kMenuBeginString) != 0)
    throw 11712;
  for (;;)
  {
    if (!inStream.GetSimpleString(s))
      throw 11704;
    if (s.CompareNoCase(kMenuEndString) == 0)
    {
      if (subMenus.Size() == 0)
        throw 11715;
      subMenus.Back()->m_ResInfo |= kMenuExFlagEnd;
      return;
    }
    subMenus.Add(new CMenuExItem());
    CMenuExItem *item = subMenus.Back();
    
    item->m_ResInfo = 0;
    item->m_Type = 0;
    item->m_State = 0;
    item->m_ID = 0;
    item->m_HelpID = 0;

    if (s.CompareNoCase(kMenuPopUpString) == 0)
      item->m_ResInfo |= kMenuExFlagPopUp;
    else if (s.CompareNoCase(kMenuItemString) != 0)
      throw 242305;
    
    UString caption;
    bool quoted;
    if (!inStream.GetString(quoted, caption))
      throw 242039;
    if (!quoted)
    {
      if (item->IsPopup())
        throw 2423066;
      if (caption.CompareNoCase(kMenuExSeparatorString) != 0)
        throw 242307;
      item->m_Type |= MFT_SEPARATOR;
      continue; // test it;
    }

    item->m_Text = g_MenuStringSubstituter.ConvertTextToBin(caption);

    if (!inStream.WaitComma())
      throw 242235;
    if (!inStream.GetSimpleString(s))
      throw 242040;
    UINT16 id;
    if (!UStringToWORD(s, id)) // Microsoft BUGS
      throw 242041;
    item->m_ID = id;


    if (inStream.WaitComma())
    {
      if (!inStream.WaitCommaNotStepOver())
      {
        for (;;)
        {
          if (!inStream.GetSimpleString(s))
            throw 242236;
          UINT32 flag;
          if (g_MenuExTypeFlags.GetOptionFlag(s, flag))
            item->m_Type |= flag;
          else if (UStringToDWORD(s, flag))
            item->m_Type |= flag;
          else
            throw 12216;
          if (!inStream.WaitOrDelimiter())
            break;
        }
      }
      if (inStream.WaitComma())
      {
        if (!inStream.WaitCommaNotStepOver())
        {
          for (;;)
          {
            if (!inStream.GetSimpleString(s))
              throw 242237;
            UINT32 flag;
            if (g_MenuExStateFlags.GetOptionFlag(s, flag))
              item->m_State |= flag;
            else if (UStringToDWORD(s, flag))
              item->m_State |= flag;
            else
              throw 12217;
            if (!inStream.WaitOrDelimiter())
              break;
          }
        }
        if (inStream.WaitComma())
        {
          if (item->IsPopup())
          {
            if (!UStringToDWORD(s, item->m_HelpID))
              throw 12218;
          }
          else
            throw 242143;
        }
      }
    }
    if (item->IsPopup())
      LoadListFromText(inStream, item->m_SubMenus);
  }
}

// ----------------------------
// Save to Text


void CMenuExItem::SaveToText(UString &outStream, CIndentMaker &indentMaker)
{
  indentMaker.WriteIndent(outStream);
  if (IsPopup())
    outStream += kMenuPopUpString;
  else
    outStream += kMenuItemString;
  outStream += kCharSpace;
  if (IsSeparator())
    outStream += kMenuExSeparatorString;
  else
  {
    outStream += g_MenuStringSubstituter.ConvertBinToText(m_Text);
    outStream += kCharComma;
    outStream += kCharSpace;
    outStream += UnsignedToUString(m_ID);

    UStringVector optionNames;
    UINT32 notUsedFlags;
    g_MenuExTypeFlags.GetOptionNames(m_Type, optionNames, notUsedFlags);
    if (notUsedFlags != 0)
      optionNames.Add(UnsignedToHexUString(notUsedFlags));
    int kNumQuotes = 0;
    if (optionNames.Size() == 0)
    {
      kNumQuotes++;
      /*
      UString name;
      if (!g_MenuExTypeFlags.GetFirstZeroFlagName(name))
        throw 12244;
      optionNames.Add(name);
      */
    }
    int i;
    for (i = 0; i < optionNames.Size(); i++)
    {
      if (i == 0)
        outStream += kCharComma;
      else
        outStream += kCharORDelimeter;
      outStream += kCharSpace;
      outStream += optionNames[i];
    }
    optionNames.Clear();
    g_MenuExStateFlags.GetOptionNames(m_State, optionNames, notUsedFlags);
    if (notUsedFlags != 0)
      optionNames.Add(UnsignedToHexUString(notUsedFlags));
    if (optionNames.Size() == 0)
    {
      kNumQuotes++;
      /*
      UString name;
      if (!g_MenuExStateFlags.GetFirstZeroFlagName(name))
        throw 12244;
      optionNames.Add(name);
      */
    }
    else
    {
      for (int i = 0; i < kNumQuotes; i++)
      {
        outStream += kCharComma;
        outStream += kCharSpace;
      }
      kNumQuotes = 0;
    }
    for (i = 0; i < optionNames.Size(); i++)
    {
      if (i == 0)
        outStream += kCharComma;
      else
        outStream += kCharORDelimeter;
      outStream += kCharSpace;
      outStream += optionNames[i];
    }
    if (IsPopup())
    {
      if (m_HelpID != 0)
      {
        for (int i = 0; i < kNumQuotes; i++)
        {
          outStream += kCharComma;
          outStream += kCharSpace;
        }
        outStream += kCharComma;
        outStream += kCharSpace;
        outStream += UnsignedToUString(m_HelpID); // test it;
      }
      AddNewLine(outStream);
      SaveListToText(outStream, indentMaker, m_SubMenus);
    }
  }
  AddNewLine(outStream);
}

void CMenuExItem::SaveListToText(UString &outStream, CIndentMaker &indentMaker,
    CMenuExItemList &subMenus)
{
  indentMaker.WriteIndent(outStream);
  outStream += kMenuBeginString;
  AddNewLine(outStream);
  indentMaker.AddIndent();

  for (int i = 0; i < subMenus.Size(); i++)
    subMenus[i]->SaveToText(outStream, indentMaker);

  indentMaker.DeleteIndent();
  indentMaker.WriteIndent(outStream);
  outStream += kMenuEndString;
}


//------------------------------------
// CDecoder

void CDecoder::ReadItems(const CItemNameID &resourceName,
    const CDataInfo &resourceData,
    CSubResourcePropertiesVector &structure) const
{
  structure.Add(CSubResourceProperties());
  CSubResourceProperties &resourceProps = structure.Back();

  resourceProps.Name = resourceName;
  resourceProps.DefaultExtension = TEXT("txt");

  UINT32 resourceSize = resourceData.Size / 2;
  UINT16 *pData = (UINT16 *)((BYTE *)resourceData.Data);
  CDataInStream inStream(pData, resourceSize);
  UINT16 version = inStream.GetWORD();
  UINT16 secondWORD = inStream.GetWORD();
  CIndentMaker indentMaker;
  UString outStream;
  if (version == 0 && secondWORD == 0)
  {
    CStandardMenuItemVector items;
    CStandardMenuItem::LoadListFromBinary(inStream, items);
    outStream += kStandardMenuString;
    AddNewLine(outStream);
    CStandardMenuItem::SaveListToText(outStream, indentMaker, items);
  }
  else if (version == 1 && secondWORD == 4)
  {
    if (resourceSize < 4)
      throw CDecoderException();
    UINT32 helpId = inStream.GetDWORD();
    if (helpId != 0)
      throw CDecoderException();
    CMenuExItemList items;
    CMenuExItem::LoadListFromBinary(inStream, items);
    outStream += kMenuExString;
    AddNewLine(outStream);
    CMenuExItem::SaveListToText(outStream, indentMaker, items);
  }
  else
    throw CDecoderException();

  resourceProps.Size = outStream.Length() * sizeof(wchar_t);
  resourceProps.Data.EnsureCapacity(resourceProps.Size);
  memcpy(resourceProps.Data, (LPCWSTR)outStream, resourceProps.Size);
}



//------------------------------------
// CEncoder

void CEncoder::GetProperties(CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_MENU);
}

void CEncoder::WriteItems(const CSubResourcePropertiesVector &structure,
    CDataBlock &resourceData) const
{
  if (structure.Size() != 1)
    throw 2053;
  const CSubResourceProperties &subResourceProps = structure[0];

  CTextInStream inStream;
  if ((subResourceProps.Size & 1) != 0)
    throw 242243;
  inStream.Init((const UINT16 *)((const BYTE *)subResourceProps.Data),
    subResourceProps.Size / sizeof(wchar_t));
  UString s;
  if (!inStream.GetSimpleString(s))
    throw 20128;
  CDataOutStream outStream;
  if (s.CompareNoCase(kStandardMenuString) == 0)
  {
    outStream.PutWORD(0); // test it
    outStream.PutWORD(0);
    CStandardMenuItemVector items;
    CStandardMenuItem::LoadListFromText(inStream, items);
    CStandardMenuItem::SaveListToBinary(outStream, items);
  }
  else if (s.CompareNoCase(kMenuExString) == 0)
  {
    outStream.PutWORD(1); // test it
    outStream.PutWORD(4);
    outStream.PutDWORD(0); // HelpID;
    CMenuExItemList items;
    CMenuExItem::LoadListFromText(inStream, items);
    CMenuExItem::SaveListToBinary(outStream, items);
  }
  else
    throw 20131;
  resourceData.Size = outStream.m_Data.Size() * sizeof(UInt16);
  resourceData.Data.EnsureCapacity(resourceData.Size);

  memcpy(resourceData.Data, &outStream.m_Data.Front(), resourceData.Size);
}

}}}
