// ResString.cpp

#include "StdAfx.h"

#include "Common/StringConvert.h"

#include "ResString.h"
#include "StringSubstitute.h"

namespace NWindows {
namespace NResource {
namespace NString {

void CDecoder::GetProperties(
    CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_STRING);
}

static const char *kStringResourceMustBeNumber = "String resource must have an integer identifier";
static const char *kTooBigIntegerIdentifier = "Too big integer identifier";

static const CSubstituteInitPair kStringSubstituteInitPairs[] =
{
  { L"\xA", L"\\n" },
  { L"\xD", L"\\r" },
  { L"\x9", L"\\t" }
};

struct CStringStringSubstituter
{
  UString ConvertBinToText(const UString &s);
};

UString CStringStringSubstituter::ConvertBinToText(const UString &s)
{
  UString result = s;
  for (int i = 0; i < sizeof(kStringSubstituteInitPairs) / sizeof(kStringSubstituteInitPairs[0]); i++)
  {
    const CSubstituteInitPair &pair = kStringSubstituteInitPairs[i];
    result.Replace(pair.Strings[0], pair.Strings[1]);
  }
  return result;
}

static CStringStringSubstituter g_StringStringSubstituter;

static const UInt32 kBunchSize = 0x10;

void CDecoder::ReadItems(const CItemNameID &resourceName,
    const CDataInfo &resourceData, CSubResourcePropertiesVector &structure) const
{
  if (!resourceName.IsNumber())
    throw kStringResourceMustBeNumber;
  
  DWORD number = resourceName.GetAsNumber();
  if (number < 1)
    throw 31759;
  
  DWORD idBase = (DWORD(number) - 1) * kBunchSize;

  WORD *wordPointer = (WORD*)(resourceData.Data);
  DWORD sumSize = 0;
  for (DWORD i = 0; i < kBunchSize; i++)
  {
    WORD size = *wordPointer;
    wordPointer++;
    sumSize++;
    if (size != 0)
    {
      DWORD nameNumber = idBase + i;
      if (nameNumber > 0xFFFF)
        throw kTooBigIntegerIdentifier;
      structure.Add(CSubResourceProperties());
      CSubResourceProperties &resourceProps = structure.Back();

      resourceProps.Name.Set(MAKEINTRESOURCE(nameNumber));
      resourceProps.Size = size * 2; // Test it;
      resourceProps.DefaultExtension = TEXT("txt");

      CByteDynamicBuffer &byteBuffer = resourceProps.Data;
      byteBuffer.EnsureCapacity(resourceProps.Size);
      memcpy(byteBuffer, wordPointer, resourceProps.Size);

      UString s;
      for (DWORD i = 0; i < size; i++)
        s += wordPointer[i];
      UString convertedResourceString = g_StringStringSubstituter.ConvertBinToText(s);
      // DWORD convertedSize = convertedResourceString.Length();

      resourceProps.Description = GetSystemString(convertedResourceString);
      // resourceProps.Description.Replace(kCharLF, kLFReplace);
      wordPointer += size;
      sumSize += size;
    }
  }
  if (sumSize * 2 != resourceData.Size)
  {
    // ShowErrorMessage("Incorrect Size of strings"); // Delphi bug support
    // return false;
  }
}

void CEncoder::GetProperties(CCoderProperties &props) const
{
  props.AllTypesAreSuported = false;
  props.SupportedType.Set(RT_STRING);
}

void CEncoder::WriteItems(const CSubResourcePropertiesVector &structure,
    CDataBlock &resourceData) const
{
  WORD mainBase = 0;
  bool baseDefined = false;
  CRecordVector<UInt16> items;

  for (UInt32 i = 0; i < kBunchSize; i++)
  {
    int j;
    for (j = 0; j < structure.Size(); j++)
    {
      const CSubResourceProperties &subResourceProps = structure[j];
      if (!subResourceProps.Name.IsNumber())
        throw 251746;
      WORD index = subResourceProps.Name.GetAsNumber();
      WORD base = (index / kBunchSize) + 1;
      if (baseDefined)
      {
        if (base != mainBase)
          throw 251749;
      }
      else
        mainBase = base;
      if ((index % kBunchSize) != i)
        continue;
      DWORD size = subResourceProps.Size;
      if ((size % 2) != 0)
        throw 251759;
      size /= 2;
      if (size >= 0x10000)
        throw 251800;
      items.Add((UInt16)size);
      const UInt16 *ptr = (const UInt16 *)((const BYTE *)subResourceProps.Data);
      for (DWORD k = 0; k < size; k++)
        items.Add(ptr[k]);
      break;
    }
    if (j == structure.Size())
      items.Add(0);
  }
  resourceData.Size = items.Size() * sizeof(UInt16);
  resourceData.Data.EnsureCapacity(resourceData.Size);
  memcpy(resourceData.Data, &items.Front(), resourceData.Size);
}

}}}
