// Windows/Resource/Utils.cpp

#include "StdAfx.h"

#include "Common/IntToString.h"
#include "Common/StringToInt.h"

#include "Windows/Defs.h"

#include "CoderUtils.h"
#include "ResUtils.h"

namespace NWindows {
namespace NResource {

using namespace std;

class CItemInfoDatabaseSpec
{
public:
  std::map<LANGID, int> LangToIndexMap;
  std::map<CItemTypeID, int> TypeIDToIndexMap;
  CObjectVector<CItemInfo> Items;
};

static const int kMaxNumDigits = 5;

bool StringToUINT16(const CSysString &value, UINT16 &result)
{
  UInt64 res64;
  const TCHAR *end;
  res64 = ConvertStringToUInt64(value, &end);
  if (*end != 0)
    return false;
  result = (UINT16)res64;
  return (res64 < (1 << 16));
}

static bool IsPointerAWORD(LPCTSTR type)
{
  return (((UINT_PTR)(type)) < 0x10000ui64); // Test it!
}


// ---------------------------------------

bool operator <(const CNumberOrString &a1, const CNumberOrString &a2)
{
  if (a1.IsNumber())
  {
    if (a2.IsNumber())
      return a1.m_Number < a2.m_Number;
    else
      return true;
  }
  if (a2.IsNumber())
    return false;
  else
    return a1.m_String.Compare(a2.m_String) < 0;
}

bool operator ==(const CNumberOrString &a1, const CNumberOrString &a2)
{
  if (a1.IsNumber())
  {
    if (a2.IsNumber())
      return a1.m_Number == a2.m_Number;
    else
      return false;
  }
  if (a2.IsNumber())
    return false;
  else
    return a1.m_String.Compare(a2.m_String) == 0;
}

bool operator !=(const CNumberOrString &a1, const CNumberOrString &a2)
{
  return !(a1 == a2);
}

/////////////////////////////////////////////////
// CNumberOrString

void CNumberOrString::Set(LPCTSTR s)
{
  m_IsNumber = IsPointerAWORD(s);
  if (m_IsNumber)
    m_Number = UINT16(s);
  else
    m_String = s;
}

void CNumberOrString::ParseAndSet(LPCTSTR s)
{
  if (StringToUINT16(s, m_Number))
    m_IsNumber = true;
  else
    SetAsName(s);
}

void CNumberOrString::SetAsNumber(UINT16 number)
{
  m_IsNumber = true;
  m_Number = number;
}

void CNumberOrString::SetAsName(LPCTSTR s)
{
  m_IsNumber = false;
  m_String = s;
}


static CSysString NormalizeFileName(const CSysString &s)
{
  if (s.Length() > 2 && s[0] == '"'  && s[s.Length() - 1] == '"' )
    return s.Mid(1, s.Length() - 2);
  return s;
}

CSysString CNumberOrString::GetAsString() const
{
  return (m_IsNumber) ? DWORDToString(m_Number) : NormalizeFileName(m_String);
}

UINT16 CNumberOrString::GetAsNumber() const
{
  if (!m_IsNumber)
    throw 31756;
  return m_Number;
};

LPCTSTR CNumberOrString::GetAsLPCTSTR() const
{
  LPCTSTR l1 = MAKEINTRESOURCE(m_Number);
  LPCTSTR l2 = m_String;
  return ((m_IsNumber) ? l1 : l2);
}


//-----------------------------------
// CItemInfoDatabaseSpec

void CItemInfoDatabase::Clear()
{
  LangIDs.Clear();
  TypeIDs.Clear();
  Items.Clear();
}

static BOOL CALLBACK EnumResLangProc(HMODULE /* module */, LPCTSTR type,
    LPCTSTR name, WORD language, LONG_PTR param)
{
  CItemInfo itemInfo;
  itemInfo.Name.Set(name);
  CItemInfoDatabaseSpec *db = (CItemInfoDatabaseSpec *)param;
  {
    CItemTypeID typeID;
    typeID.Set(type);
    map<CItemTypeID, int>::const_iterator iterator =
        db->TypeIDToIndexMap.find(typeID);
    if (iterator == db->TypeIDToIndexMap.end())
    {
      itemInfo.TypeIndex = db->TypeIDToIndexMap.size();
      db->TypeIDToIndexMap.insert(
          map<CItemTypeID, int>::value_type(typeID, itemInfo.TypeIndex));
    }
    else
      itemInfo.TypeIndex = (*iterator).second;
  }

  {
    map<LANGID, int>::const_iterator iterator =
        db->LangToIndexMap.find(language);
    if (iterator == db->LangToIndexMap.end())
    {
      itemInfo.LangIndex = db->LangToIndexMap.size();
      db->LangToIndexMap.insert(
          map<LANGID, int>::value_type(language, itemInfo.LangIndex));
    }
    else
      itemInfo.LangIndex = (*iterator).second;
  }
  db->Items.Add(itemInfo);
  return TRUE;
}

static BOOL CALLBACK EnumResNameProc(HMODULE module, LPCTSTR type, LPTSTR name, LONG param)
{
  return ::EnumResourceLanguages(module, type, name, EnumResLangProc, param);
}

static BOOL CALLBACK EnumResTypeProc(HMODULE module, LPTSTR type, LONG param)
{
  return EnumResourceNames(module, type, EnumResNameProc, param);
}


// -----------------------------------
// CModule

bool CModule::ReadItemInfoDatabase(CItemInfoDatabase &db) const
{
  db.Clear();
  CItemInfoDatabaseSpec databaseSpec;
  if (!BOOLToBool(::EnumResourceTypes(_module, &EnumResTypeProc, LONG_PTR(&databaseSpec))))
    return false;
  db.Items = databaseSpec.Items;
  {
    for (map<CItemTypeID, int>::const_iterator i = databaseSpec.TypeIDToIndexMap.begin();
        i != databaseSpec.TypeIDToIndexMap.end(); i++)
    {
      int index = (*i).second;
      for (int j = db.TypeIDs.Size(); j <= index; j++)
        db.TypeIDs.Add(CItemTypeID());
      db.TypeIDs[index] = (*i).first;
    }
  }
  {
    for (map<LANGID, int>::const_iterator i = databaseSpec.LangToIndexMap.begin();
        i != databaseSpec.LangToIndexMap.end(); i++)
    {
      int index = (*i).second;
      for (int j = db.LangIDs.Size(); j <= index; j++)
        db.LangIDs.Add(0);
      db.LangIDs[index] = (*i).first;
    }
  }
  return true;
}

bool CModule::ReadResource(CDataInfo &resourceData, const CItemInfoDatabase &db, int index) const
{
  const CItemInfo &itemInfo = db.Items[index];

  resourceData.Data = NULL;
  resourceData.Size = 0;

  HRSRC aHRSRC = ::FindResourceEx(_module,
      db.TypeIDs[itemInfo.TypeIndex].GetAsLPCTSTR(),
      itemInfo.Name.GetAsLPCTSTR(), db.LangIDs[itemInfo.LangIndex]);
  if (aHRSRC == NULL)
    return false;
  resourceData.Size = ::SizeofResource(_module, aHRSRC);
  HGLOBAL aHGLOBAL = ::LoadResource(_module, aHRSRC);
  if (aHGLOBAL == NULL)
    return false;
  resourceData.Data = ::LockResource(aHGLOBAL);
  return (resourceData.Data != NULL);
}
  
//////////////////////////////////
// CStandardTypesNames

void CStandardTypesNames::AddType(LPCTSTR number, const CSysString &name)
{
  UINT16 number2 = (UINT16)number;
  m_NumberToStringMap.insert(map<UINT16, CSysString>::value_type(number2, name));
  m_StringToNumberMap.insert(map<CSysString, UINT16>::value_type(name, number2));
}

bool CStandardTypesNames::GetTypeName(UINT16 number, CSysString &name) const
{
  map<UINT16, CSysString>::const_iterator iterator = m_NumberToStringMap.find(number);
  if (iterator != m_NumberToStringMap.end())
  {
    name = (*iterator).second;
    return true;
  }
  else
    return false;
};

bool CStandardTypesNames::GetTypeNumber(const CSysString &name, UINT16 &number) const
{
  map<CSysString, UINT16>::const_iterator iterator = m_StringToNumberMap.find(name);
  if (iterator != m_StringToNumberMap.end())
  {
    number = (*iterator).second;
    return true;
  }
  else
    return false;
};

void CStandardTypesNames::AddEnglishNames()
{
  AddType(RT_CURSOR, TEXT("CURSOR"));
  AddType(RT_BITMAP, TEXT("BITMAP"));
  AddType(RT_ICON, TEXT("ICON"));
  AddType(RT_MENU, TEXT("MENU"));
  AddType(RT_DIALOG, TEXT("DIALOG"));
  AddType(RT_STRING, TEXT("STRING"));
  AddType(RT_FONTDIR, TEXT("FONTDIR"));
  AddType(RT_FONT, TEXT("FONT"));
  AddType(RT_ACCELERATOR, TEXT("ACCELERATOR"));
  AddType(RT_RCDATA, TEXT("RCDATA"));
  AddType(RT_MESSAGETABLE, TEXT("MESSAGETABLE"));
  
  AddType(RT_GROUP_CURSOR, TEXT("GROUP_CURSOR"));
  AddType(RT_GROUP_ICON, TEXT("GROUP_ICON"));
  AddType(RT_VERSION, TEXT("VERSION"));
  AddType(RT_DLGINCLUDE, TEXT("DLGINCLUDE"));

  AddType(RT_PLUGPLAY, TEXT("PLUGPLAY"));
  AddType(RT_VXD, TEXT("VXD"));
  AddType(RT_ANICURSOR, TEXT("ANICURSOR"));
  AddType(RT_ANIICON, TEXT("ANIICON"));
  AddType(RT_HTML, TEXT("HTML"));
  AddType(RT_MANIFEST, TEXT("MANIFEST"));
}

}}
